@@ -15,7 +15,7 @@
  #
  # Originally developed by Gregory Trubetskoy.
  #
- # $Id: Cookie.py 102649 2004-02-16 19:47:28Z grisha $
+ # $Id: Cookie.py 374268 2006-02-02 05:31:45Z nlehuen $
 
 """
 
@@ -46,7 +46,7 @@ import hmac
 import marshal
 import base64
 
-import apache
+# import apache
 
 class CookieError(Exception):
     pass
@@ -260,7 +260,10 @@ class MarshalCookie(SignedCookie):
 
     def __str__(self):
         
-        m = base64.encodestring(marshal.dumps(self.value))[:-1]
+        m = base64.encodestring(marshal.dumps(self.value))
+        # on long cookies, the base64 encoding can contain multiple lines
+        # separated by \n or \r\n
+        m = ''.join(m.split())
 
         result = ["%s=%s%s" % (self.name, self.hexdigest(m), m)]
         for name in self._valid_attr:
@@ -297,41 +300,19 @@ _cookiePattern = re.compile(
     )
 
 def _parse_cookie(str, Class):
-
     # XXX problem is we should allow duplicate
     # strings
     result = {}
 
-    # max-age is a problem because of the '-'
-    # XXX there should be a more elegant way
-    valid = Cookie._valid_attr + ("max-age",)
-
-    c = None
     matchIter = _cookiePattern.finditer(str)
 
     for match in matchIter:
-
         key, val = match.group("key"), match.group("val")
 
-        if not c:
-            # new cookie
-            c = Class(key, val)
-            result[key] = c
-
-        l_key = key.lower()
-        
-        if (l_key in valid or key[0] == '$'):
-            
-            # "internal" attribute, add to cookie
-
-            if l_key == "max-age":
-                l_key = "max_age"
-            setattr(c, l_key, val)
-
-        else:
-            # start a new cookie
-            c = Class(l_key, val)
-            result[l_key] = c
+        # We just ditch the cookies names which start with a dollar sign since
+        # those are in fact RFC2965 cookies attributes. See bug [#MODPYTHON-3].
+        if key[0]!='$':
+            result[key] = Class(key, val)
 
     return result
 
