@@ -18,7 +18,7 @@
  *
  * _apachemodule.c 
  *
- * $Id: _apachemodule.c 102649 2004-02-16 19:47:28Z grisha $
+ * $Id: _apachemodule.c 332807 2005-11-12 18:11:14Z nlehuen $
  *
  */
 
@@ -380,6 +380,14 @@ static PyObject *_global_lock(PyObject *self, PyObject *args)
     apr_pool_userdata_get((void **)&glb, MP_CONFIG_KEY,
                           s->process->pool);
     
+    if ((index >= (glb->nlocks)) || (index < -1)) {
+        ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
+                     "Index %d is out of range for number of global mutex locks", index);
+        PyErr_SetString(PyExc_ValueError,
+                        "Lock index is out of range for number of global mutex locks");
+        return NULL;
+    }
+    
     if (index == -1) {
 
         int hash = PyObject_Hash(key);
@@ -417,6 +425,89 @@ static PyObject *_global_lock(PyObject *self, PyObject *args)
     return Py_None;
 }
 
+/**
+ ** _global_trylock
+ **
+ *   Try to lock one of our global_mutexes
+ */
+
+static PyObject *_global_trylock(PyObject *self, PyObject *args)
+{
+
+    PyObject *server;
+    PyObject *key;
+    server_rec *s;
+    py_global_config *glb;
+    int index = -1;
+    apr_status_t rv;
+
+    if (! PyArg_ParseTuple(args, "OO|i", &server, &key, &index)) 
+        return NULL;
+
+    if (!  MpServer_Check(server)) {
+        PyErr_SetString(PyExc_TypeError,
+                        "First argument must be a server object");
+        return NULL;
+    }
+
+    s = ((serverobject *)server)->server;
+
+    apr_pool_userdata_get((void **)&glb, MP_CONFIG_KEY,
+                          s->process->pool);
+
+    if ((index >= (glb->nlocks)) || (index < -1)) {
+        ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
+                     "Index %d is out of range for number of global mutex locks", index);
+        PyErr_SetString(PyExc_ValueError,
+                        "Lock index is out of range for number of global mutex locks");
+        return NULL;
+    }
+   
+    if (index == -1) {
+
+        int hash = PyObject_Hash(key);
+        if (hash == -1) {
+            return NULL;
+        }
+        else {
+            hash = abs(hash);
+        }
+
+        /* note that this will never result in 0,
+         * which is reserved for things like dbm
+         * locking (see Session.py)
+         */
+        
+        index = (hash % (glb->nlocks-1)+1);
+    }
+
+//      ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
+//                "_global_trylock at index %d from pid %d", index, getpid());
+    Py_BEGIN_ALLOW_THREADS
+    rv = apr_global_mutex_trylock(glb->g_locks[index]);        
+    Py_END_ALLOW_THREADS
+    
+    if (rv == APR_SUCCESS) {
+//      ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
+//                "_global_trylock DONE at index %d from pid %d", index, getpid());
+        Py_INCREF(Py_True);
+        return Py_True;
+    }
+    else if(APR_STATUS_IS_EBUSY(rv)) {
+//      ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
+//                "_global_trylock BUSY at index %d from pid %d", index, getpid());
+        Py_INCREF(Py_False);
+        return Py_False;
+    }
+    else {
+        ap_log_error(APLOG_MARK, APLOG_WARNING, rv, s,
+                     "Failed to acquire global mutex lock at index %d", index);
+        PyErr_SetString(PyExc_ValueError,
+                        "Failed to acquire global mutex lock");
+        return NULL;
+    }
+}
+
 /**
  ** _global_unlock
  **
@@ -446,7 +537,15 @@ static PyObject *_global_unlock(PyObject *self, PyObject *args)
 
     apr_pool_userdata_get((void **)&glb, MP_CONFIG_KEY,
                           s->process->pool);
-    
+
+    if ((index >= (glb->nlocks)) || (index < -1)) {
+        ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
+                     "Index %d is out of range for number of global mutex locks", index);
+        PyErr_SetString(PyExc_ValueError,
+                        "Lock index is out of range for number of global mutex locks");
+        return NULL;
+    }
+   
     if (index == -1) {
 
         int hash = PyObject_Hash(key);
@@ -500,16 +599,97 @@ static PyObject *mpm_query(PyObject *self, PyObject *code)
     return PyInt_FromLong(result);
 }
 
+/**
+ ** register_cleanup(interpreter, server, handler, data)
+ **
+ *    more low level version of request.register_cleanup where it is
+ *    necessary to specify the actual interpreter name. the server pool
+ *    is used. the server pool gets destroyed before the child dies or
+ *    when the whole process dies in multithreaded situations.
+ */
+
+static PyObject *register_cleanup(PyObject *self, PyObject *args)
+{
+
+    cleanup_info *ci;
+    char *interpreter = NULL;
+    serverobject *server = NULL;
+    PyObject *handler = NULL;
+    PyObject *data = NULL;
+
+    if (! PyArg_ParseTuple(args, "sOO|O", &interpreter, &server, &handler, &data))
+        return NULL; 
+
+    if (! MpServer_Check(server)) {
+        PyErr_SetString(PyExc_ValueError, 
+                        "second argument must be a server object");
+        return NULL;
+    }
+    else if(!PyCallable_Check(handler)) {
+        PyErr_SetString(PyExc_ValueError, 
+                        "third argument must be a callable object");
+        return NULL;
+    }
+    
+    ci = (cleanup_info *)malloc(sizeof(cleanup_info));
+    ci->request_rec = NULL;
+    ci->server_rec = server->server;
+    Py_INCREF(handler);
+    ci->handler = handler;
+    ci->interpreter = strdup(interpreter);
+    if (data) {
+        Py_INCREF(data);
+        ci->data = data;
+    }
+    else {
+        Py_INCREF(Py_None);
+        ci->data = Py_None;
+    }
+    
+    apr_pool_cleanup_register(child_init_pool, ci, python_cleanup, 
+                              apr_pool_cleanup_null);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+/**
+ ** exists_config_define(name)
+ **
+ *  Check for a definition from the server command line.
+ */
+
+static PyObject *exists_config_define(PyObject *self, PyObject *args)
+{
+
+    char *name = NULL;
+
+    if (! PyArg_ParseTuple(args, "s", &name))
+        return NULL; 
+
+    if(ap_exists_config_define(name)) {
+        Py_INCREF(Py_True);
+        return Py_True;
+    }
+    else {
+        Py_INCREF(Py_False);
+        return Py_False;
+    }
+}
+
 /* methods of _apache */
 struct PyMethodDef _apache_module_methods[] = {
-    {"config_tree",               (PyCFunction)config_tree,      METH_NOARGS},
-    {"log_error",                 (PyCFunction)mp_log_error,     METH_VARARGS},
-    {"mpm_query",                 (PyCFunction)mpm_query,        METH_O},
-    {"parse_qs",                  (PyCFunction)parse_qs,         METH_VARARGS},
-    {"parse_qsl",                 (PyCFunction)parse_qsl,        METH_VARARGS},
-    {"server_root",               (PyCFunction)server_root,      METH_NOARGS},
-    {"_global_lock",              (PyCFunction)_global_lock,     METH_VARARGS},
-    {"_global_unlock",            (PyCFunction)_global_unlock,   METH_VARARGS},
+    {"config_tree",           (PyCFunction)config_tree,          METH_NOARGS},
+    {"log_error",             (PyCFunction)mp_log_error,         METH_VARARGS},
+    {"mpm_query",             (PyCFunction)mpm_query,            METH_O},
+    {"parse_qs",              (PyCFunction)parse_qs,             METH_VARARGS},
+    {"parse_qsl",             (PyCFunction)parse_qsl,            METH_VARARGS},
+    {"server_root",           (PyCFunction)server_root,          METH_NOARGS},
+    {"register_cleanup",      (PyCFunction)register_cleanup,     METH_VARARGS},
+    {"exists_config_define",  (PyCFunction)exists_config_define, METH_VARARGS},
+    {"_global_lock",          (PyCFunction)_global_lock,         METH_VARARGS},
+    {"_global_trylock",       (PyCFunction)_global_trylock,      METH_VARARGS},
+    {"_global_unlock",        (PyCFunction)_global_unlock,       METH_VARARGS},
     {NULL, NULL} /* sentinel */
 };
 
