@@ -18,7 +18,7 @@
  *
  * connobject.c 
  *
- * $Id: connobject.c 102649 2004-02-16 19:47:28Z grisha $
+ * $Id: connobject.c 374510 2006-02-02 21:30:55Z jgallacher $
  *
  */
 
@@ -74,14 +74,16 @@ static PyObject * _conn_read(conn_rec *c, ap_input_mode_t mode, long len)
 
     bufsize = len == 0 ? HUGE_STRING_LEN : len;
 
-    Py_BEGIN_ALLOW_THREADS;
-    rc = ap_get_brigade(c->input_filters, bb, mode, APR_BLOCK_READ, bufsize);
-    Py_END_ALLOW_THREADS;
+    while (APR_BRIGADE_EMPTY(bb)) {
+        Py_BEGIN_ALLOW_THREADS;
+        rc = ap_get_brigade(c->input_filters, bb, mode, APR_BLOCK_READ, bufsize);
+        Py_END_ALLOW_THREADS;
 
-    if (! APR_STATUS_IS_SUCCESS(rc)) {
-        PyErr_SetObject(PyExc_IOError, 
-                        PyString_FromString("Connection read error"));
-        return NULL;
+        if (! APR_STATUS_IS_SUCCESS(rc)) {
+            PyErr_SetObject(PyExc_IOError, 
+                            PyString_FromString("Connection read error"));
+            return NULL;
+        }
     }
 
     /* 
@@ -135,7 +137,7 @@ static PyObject * _conn_read(conn_rec *c, ap_input_mode_t mode, long len)
 
             _PyString_Resize(&result, bufsize + HUGE_STRING_LEN);
             buffer = PyString_AS_STRING((PyStringObject *) result);
-            buffer += HUGE_STRING_LEN;
+            buffer += bufsize;
             bufsize += HUGE_STRING_LEN;
         }
 
@@ -258,8 +260,8 @@ static struct memberlist conn_memberlist[] = {
     {"keepalive",          T_INT,       0,                       RO},
     {"double_reverse",     T_INT,       0,                       RO},
     {"keepalives",         T_INT,       OFF(keepalives),         RO},
-    {"local_ip",           T_STRING,    OFF(remote_ip),          RO},
-    {"local_host",         T_STRING,    OFF(remote_host),        RO},
+    {"local_ip",           T_STRING,    OFF(local_ip),          RO},
+    {"local_host",         T_STRING,    OFF(local_host),        RO},
     {"id",                 T_LONG,      OFF(id),                 RO},
     /* XXX conn_config? */
     {"notes",              T_OBJECT,    0,                       RO},
@@ -292,12 +294,19 @@ static void conn_dealloc(connobject *self)
 
 static PyObject *makeipaddr(struct apr_sockaddr_t *addr)
 {
-    long x = ntohl(addr->sa.sin.sin_addr.s_addr);
-    char buf[100];
-    sprintf(buf, "%d.%d.%d.%d",
-            (int) (x>>24) & 0xff, (int) (x>>16) & 0xff,
-            (int) (x>> 8) & 0xff, (int) (x>> 0) & 0xff);
-    return PyString_FromString(buf);
+    char *str = NULL;
+    apr_status_t rc;
+    PyObject *ret = NULL;
+
+    rc = apr_sockaddr_ip_get( &str, addr );
+    if (rc==APR_SUCCESS) {
+        ret = PyString_FromString( str );
+    }
+    else {
+        PyErr_SetString(PyExc_SystemError,"apr_sockaddr_ip_get failure");
+    }
+    
+    return ret; 
 }
 
 /**
@@ -311,7 +320,13 @@ static PyObject *makesockaddr(struct apr_sockaddr_t *addr)
     PyObject *addrobj = makeipaddr(addr);
     PyObject *ret = NULL;
     if (addrobj) {
-        ret = Py_BuildValue("Oi", addrobj, ntohs(addr->sa.sin.sin_port));
+        apr_port_t port;
+        if(apr_sockaddr_port_get(&port, addr)==APR_SUCCESS) {
+            ret = Py_BuildValue("Oi", addrobj, port );
+        }
+        else {
+            PyErr_SetString(PyExc_SystemError,"apr_sockaddr_port_get failure");
+        }
         Py_DECREF(addrobj);
     }
     return ret;
