@@ -18,7 +18,7 @@
  *
  * mod_python.c 
  *
- * $Id: mod_python.c 102649 2004-02-16 19:47:28Z grisha $
+ * $Id: mod_python.c 367835 2006-01-10 23:36:37Z jgallacher $
  *
  * See accompanying documentation and source code comments 
  * for details.
@@ -31,12 +31,16 @@
  * (In a Python dictionary) */
 static PyObject * interpreters = NULL;
 
+#if APR_HAS_THREADS
+static apr_thread_mutex_t* interpreters_lock = 0;
+#endif
+
 apr_pool_t *child_init_pool = NULL;
 
 /**
  ** make_interpreter
  **
- *      Creates a new Python interpeter.
+ *      Creates a new Python interpreter.
  */
 
 static PyInterpreterState *make_interpreter(const char *name, server_rec *srv)
@@ -73,7 +77,7 @@ static PyInterpreterState *make_interpreter(const char *name, server_rec *srv)
  *      the reference to obCallBack.
  */
 
-static PyObject * make_obcallback(server_rec *s)
+static PyObject * make_obcallback(char *name, server_rec *s)
 {
 
     PyObject *m;
@@ -94,14 +98,16 @@ static PyObject * make_obcallback(server_rec *s)
 
     if (! ((m = PyImport_ImportModule(MODULENAME)))) {
         ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, s,
-                     "make_obcallback: could not import %s.\n", MODULENAME);
+                     "make_obcallback: could not import %s.\n", (!MODULENAME) ? "<null>" : MODULENAME);
         PyErr_Print();
+        fflush(stderr); 
     }
     
-    if (m && ! ((obCallBack = PyObject_CallMethod(m, INITFUNC, NULL)))) {
+    if (m && ! ((obCallBack = PyObject_CallMethod(m, INITFUNC, "sO", name, MpServer_FromServer(s))))) {
         ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, s,
-                     "make_obcallback: could not call %s.\n", INITFUNC);
+                     "make_obcallback: could not call %s.\n", (!INITFUNC) ? "<null>" : INITFUNC);
         PyErr_Print();
+        fflush(stderr); 
     }
     
     return obCallBack;
@@ -123,12 +129,18 @@ static interpreterdata *get_interpreter(const char *name, server_rec *srv)
     if (! name)
         name = MAIN_INTERPRETER;
 
+#if APR_HAS_THREADS
+    apr_thread_mutex_lock(interpreters_lock);
+#endif
 #ifdef WITH_THREAD
     PyEval_AcquireLock();
 #endif
 
-    if (!interpreters)
+    if (!interpreters) {
+         ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, srv,
+                       "python_handler: interpreters dictionary not initialised.");
         return NULL;
+    }
 
     p = PyDict_GetItemString(interpreters, (char *)name);
     if (!p)
@@ -141,6 +153,7 @@ static interpreterdata *get_interpreter(const char *name, server_rec *srv)
             idata->obcallback = NULL; 
             p = PyCObject_FromVoidPtr((void *) idata, NULL);
             PyDict_SetItemString(interpreters, (char *)name, p);
+            Py_DECREF(p);
         }
     }
     else {
@@ -150,6 +163,9 @@ static interpreterdata *get_interpreter(const char *name, server_rec *srv)
 #ifdef WITH_THREAD
     PyEval_ReleaseLock();
 #endif
+#if APR_HAS_THREADS
+    apr_thread_mutex_unlock(interpreters_lock);
+#endif
 
     if (! idata) {
         ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, srv,
@@ -167,14 +183,18 @@ static interpreterdata *get_interpreter(const char *name, server_rec *srv)
 
     if (!idata->obcallback) {
 
-        idata->obcallback = make_obcallback(srv);
+        idata->obcallback = make_obcallback((char*)name,srv);
 
         if (!idata->obcallback) 
         {
 #ifdef WITH_THREAD
             PyEval_ReleaseThread(tstate);
+#else
+            PyThreadState_Swap(NULL);
 #endif
             PyThreadState_Delete(tstate);
+            ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, srv,
+                      "python_handler: no interpreter callback found.");
             return NULL;
         }
     }
@@ -195,6 +215,8 @@ static void release_interpreter(void)
     PyThreadState *tstate = PyThreadState_Get();
 #ifdef WITH_THREAD
     PyEval_ReleaseThread(tstate);
+#else
+    PyThreadState_Swap(NULL);
 #endif
     PyThreadState_Delete(tstate);
 }
@@ -223,6 +245,7 @@ apr_status_t python_cleanup(void *data)
     if (!idata) {
         Py_DECREF(ci->handler);
         Py_XDECREF(ci->data);
+        free(ci->interpreter);
         free(ci);
         return APR_SUCCESS; /* this is ignored anyway */
     }
@@ -469,6 +492,9 @@ static int python_init(apr_pool_t *p, apr_pool_t *ptemp,
     const char *userdata_key = "python_init";
     apr_status_t rc;
 
+    /* fudge for Mac OS X with Apache 1.3 where Py_IsInitialized() broke */
+    static int initialized = 0;
+
     apr_pool_userdata_get(&data, userdata_key, s->process->pool);
     if (!data) {
         apr_pool_userdata_set((const void *)1, userdata_key,
@@ -490,12 +516,16 @@ static int python_init(apr_pool_t *p, apr_pool_t *ptemp,
     }
 
     /* initialize global Python interpreter if necessary */
-    if (! Py_IsInitialized()) 
+    if (initialized == 0 || !Py_IsInitialized()) 
     {
-
+        initialized = 1;
+        
         /* initialze the interpreter */
         Py_Initialize();
 
+#if APR_HAS_THREADS
+        apr_thread_mutex_create(&interpreters_lock, APR_THREAD_MUTEX_UNNESTED, p);
+#endif
 #ifdef WITH_THREAD
         /* create and acquire the interpreter lock */
         PyEval_InitThreads();
@@ -558,8 +588,8 @@ static void *python_create_dir_config(apr_pool_t *p, char *dir)
     py_config *conf = python_create_config(p);
 
     /* make sure directory ends with a slash */
-    if (dir && (dir[strlen(dir) - 1] != SLASH))
-        conf->config_dir = apr_pstrcat(p, dir, SLASH_S, NULL);
+    if (dir && (dir[strlen(dir) - 1] != '/'))
+        conf->config_dir = apr_pstrcat(p, dir, "/", NULL);
     else
         conf->config_dir = apr_pstrdup(p, dir);
 
@@ -581,16 +611,35 @@ static void *python_create_srv_config(apr_pool_t *p, server_rec *srv)
     return conf;
 }
 
+/*
+code begin
+*/
+/**
+ ** modpython_table_overlap
+ **
+ * Replaces the apr_table_overlap() function using a specific pool
+ * for the resulting table.
+ */
+
+static apr_table_t *modpython_table_overlap(apr_pool_t *p,
+                                            apr_table_t *current_table,
+                                            apr_table_t *new_table)
+{
+    apr_table_t *merge = apr_table_overlay(p, current_table, new_table);
+    apr_table_compress(merge, APR_OVERLAP_TABLES_SET);
+    return merge;
+}
+
 /**
  ** python_merge_dir_config
  **
  */
 
-static void *python_merge_config(apr_pool_t *p, void *current_conf, 
+static void *python_merge_config(apr_pool_t *p, void *current_conf,
                                  void *new_conf)
 {
 
-    py_config *merged_conf = 
+    py_config *merged_conf =
         (py_config *) apr_pcalloc(p, sizeof(py_config));
     py_config *cc = (py_config *) current_conf;
     py_config *nc = (py_config *) new_conf;
@@ -605,20 +654,19 @@ static void *python_merge_config(apr_pool_t *p, void *current_conf,
      */
 
     /** create **/
-    merged_conf->directives = apr_table_make(p, 4);
-    merged_conf->options = apr_table_make(p, 4);
     merged_conf->hlists = apr_hash_make(p);
     merged_conf->in_filters = apr_hash_make(p);
     merged_conf->out_filters = apr_hash_make(p);
 
-    /** copy current **/
+    /** merge directives and options **/
+    merged_conf->directives = modpython_table_overlap(p, cc->directives,
+                                                         nc->directives);
+    merged_conf->options = modpython_table_overlap(p, cc->options,
+                                                      nc->options);
 
+    /** copy current **/
     merged_conf->authoritative = cc->authoritative;
     merged_conf->config_dir = apr_pstrdup(p, cc->config_dir);
-    apr_table_overlap(merged_conf->directives, cc->directives,
-                      APR_OVERLAP_TABLES_SET);
-    apr_table_overlap(merged_conf->options, cc->options,
-                      APR_OVERLAP_TABLES_SET);
 
     for (hi = apr_hash_first(p, cc->hlists); hi; hi=apr_hash_next(hi)) {
         apr_hash_this(hi, (const void **)&key, &klen, (void **)&hle);
@@ -642,11 +690,6 @@ static void *python_merge_config(apr_pool_t *p, void *current_conf,
     if (nc->config_dir)
         merged_conf->config_dir = apr_pstrdup(p, nc->config_dir);
 
-    apr_table_overlap(merged_conf->directives, nc->directives,
-                      APR_OVERLAP_TABLES_SET);
-    apr_table_overlap(merged_conf->options, nc->options,
-                      APR_OVERLAP_TABLES_SET);
-
     for (hi = apr_hash_first(p, nc->hlists); hi; hi=apr_hash_next(hi)) {
         apr_hash_this(hi, (const void**)&key, &klen, (void **)&hle);
         apr_hash_set(merged_conf->hlists, key, klen, (void *)hle);
@@ -665,6 +708,11 @@ static void *python_merge_config(apr_pool_t *p, void *current_conf,
     return (void *) merged_conf;
 }
 
+/*
+code end
+*/
+
+
 /**
  ** python_directive
  **
@@ -767,7 +815,7 @@ static const char *python_directive_handler(cmd_parms *cmd, py_config* conf,
  */
 
 static const char *python_directive_flag(void * mconfig, 
-                                         char *key, int val)
+                                         char *key, int val, int set)
 {
     py_config *conf;
 
@@ -777,7 +825,12 @@ static const char *python_directive_flag(void * mconfig,
         apr_table_set(conf->directives, key, "1");
     }
     else {
-        apr_table_unset(conf->directives, key);
+        if (set) {
+            apr_table_set(conf->directives, key, "0");
+        }
+        else {
+            apr_table_unset(conf->directives, key);
+        }
     }
 
     return NULL;
@@ -806,7 +859,7 @@ static requestobject *get_request_object(request_rec *req, const char *interp_na
     }
     else {
 /*         if ((req->path_info) &&  */
-/*             (req->path_info[strlen(req->path_info) - 1] == SLASH)) */
+/*             (req->path_info[strlen(req->path_info) - 1] == '/')) */
 /*         { */
 /*             int i; */
 /*             i = strlen(req->path_info); */
@@ -821,7 +874,7 @@ static requestobject *get_request_object(request_rec *req, const char *interp_na
 /*             if (!request_obj) return NULL; */
 
 /*             /\* put the slash back in *\/ */
-/*             req->path_info[i - 1] = SLASH;  */
+/*             req->path_info[i - 1] = '/';  */
 /*             req->path_info[i] = 0; */
 
 /*             /\* and also make PATH_INFO == req->subprocess_env *\/ */
@@ -887,10 +940,21 @@ static const char *select_interp_name(request_rec *req, conn_rec *con, py_config
         if ((s = apr_table_get(conf->directives, "PythonInterpPerDirectory"))) {
             
             /* base interpreter on directory where the file is found */
-            if (req && ap_is_directory(req->pool, req->filename))
-                return ap_make_dirstr_parent(req->pool, 
-                                             apr_pstrcat(req->pool, req->filename, 
-                                                         SLASH_S, NULL ));
+            if (req && ap_is_directory(req->pool, req->filename)) {
+                /** XXX I suppose that if req->filename is a directory, there already
+                    is a trailing slash in req->filename. This is due to the fact
+                    that Apache redirect any request from /directory to /directory/.
+                    That's why the tests below are commented out, they should be useless.
+                **/
+                /* if (req->filename[strlen(req->filename)-1]=='/') { */
+                    return ap_make_dirstr_parent(req->pool, req->filename);
+                /* }
+                else {
+                    return ap_make_dirstr_parent(req->pool, 
+                                                apr_pstrcat(req->pool, req->filename, 
+                                                            "/", NULL ));
+                } */
+            }
             else {
                 if (req && req->filename)
                     return ap_make_dirstr_parent(req->pool, req->filename);
@@ -1015,8 +1079,11 @@ static int python_handler(request_rec *req, char *phase)
     /* get/create interpreter */
     idata = get_interpreter(interp_name, req->server);
 
-    if (!idata)
+    if (!idata) {
+        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, req,
+                      "python_handler: Can't get/create interpreter.");
         return HTTP_INTERNAL_SERVER_ERROR;
+    }
     
     /* create/acquire request object */
     request_obj = get_request_object(req, interp_name, phase);
@@ -1025,12 +1092,18 @@ static int python_handler(request_rec *req, char *phase)
     if (ext) 
         request_obj->extension = apr_pstrdup(req->pool, ext);
 
-    /* create a hahdler list object */
-    request_obj->hlo = (hlistobject *)MpHList_FromHLEntry(hle);
+    if (!hle) {
+        /* create a handler list object from dynamically registered handlers */
+        request_obj->hlo = (hlistobject *)MpHList_FromHLEntry(dynhle);
+    }
+    else {
+        /* create a handler list object */
+        request_obj->hlo = (hlistobject *)MpHList_FromHLEntry(hle);
 
-    /* add dynamically registered handlers, if any */
-    if (dynhle) {
-        MpHList_Append(request_obj->hlo, dynhle);
+        /* add dynamically registered handlers, if any */
+        if (dynhle) {
+            MpHList_Append(request_obj->hlo, dynhle);
+        }
     }
 
     /* 
@@ -1199,8 +1272,11 @@ static apr_status_t python_connection(conn_rec *con)
     /* get/create interpreter */
     idata = get_interpreter(interp_name, con->base_server);
 
-    if (!idata)
+    if (!idata) {
+        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, con->base_server,
+                      "python_handler: Can't get/create interpreter.");
         return HTTP_INTERNAL_SERVER_ERROR;
+    }
     
     /* create connection object */
     conn_obj = (connobject*) MpConn_FromConn(con);
@@ -1296,8 +1372,11 @@ static apr_status_t python_filter(int is_input, ap_filter_t *f,
     /* get/create interpreter */
     idata = get_interpreter(interp_name, req->server);
    
-    if (!idata)
+    if (!idata) {
+        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, req,
+                      "python_handler: Can't get/create interpreter.");
         return HTTP_INTERNAL_SERVER_ERROR;
+    }
 
     /* create/acquire request object */
     request_obj = get_request_object(req, interp_name, 
@@ -1433,13 +1512,13 @@ static const char *directive_PythonInterpreter(cmd_parms *cmd, void *mconfig,
  */
 static const char *directive_PythonDebug(cmd_parms *cmd, void *mconfig,
                                          int val) {
-    const char *rc = python_directive_flag(mconfig, "PythonDebug", val);
+    const char *rc = python_directive_flag(mconfig, "PythonDebug", val, 0);
 
     if (!rc) {
         py_config *conf = ap_get_module_config(cmd->server->module_config,
                                                &python_module);
 
-        return python_directive_flag(conf, "PythonDebug", val);
+        return python_directive_flag(conf, "PythonDebug", val, 0);
     }
     return rc;
 }
@@ -1452,12 +1531,12 @@ static const char *directive_PythonDebug(cmd_parms *cmd, void *mconfig,
  */
 static const char *directive_PythonEnablePdb(cmd_parms *cmd, void *mconfig,
                                              int val) {
-    const char *rc = python_directive_flag(mconfig, "PythonEnablePdb", val);
+    const char *rc = python_directive_flag(mconfig, "PythonEnablePdb", val, 0);
 
     if (!rc) {
         py_config *conf = ap_get_module_config(cmd->server->module_config,
                                                &python_module);
-        return python_directive_flag(conf, "PythonEnablePdb", val);
+        return python_directive_flag(conf, "PythonEnablePdb", val, 0);
     }
     return rc;
 }
@@ -1471,12 +1550,12 @@ static const char *directive_PythonEnablePdb(cmd_parms *cmd, void *mconfig,
 
 static const char *directive_PythonInterpPerDirective(cmd_parms *cmd, 
                                                       void *mconfig, int val) {
-    const char *rc = python_directive_flag(mconfig, "PythonInterpPerDirective", val);
+    const char *rc = python_directive_flag(mconfig, "PythonInterpPerDirective", val, 0);
 
     if (!rc) {
         py_config *conf = ap_get_module_config(cmd->server->module_config,
                                                &python_module);
-        return python_directive_flag(conf, "PythonInterpPerDirective", val);
+        return python_directive_flag(conf, "PythonInterpPerDirective", val, 0);
     }
     return rc;
 }
@@ -1490,7 +1569,7 @@ static const char *directive_PythonInterpPerDirective(cmd_parms *cmd,
 
 static const char *directive_PythonInterpPerDirectory(cmd_parms *cmd, 
                                                       void *mconfig, int val) {
-    return python_directive_flag(mconfig, "PythonInterpPerDirectory", val);
+    return python_directive_flag(mconfig, "PythonInterpPerDirectory", val, 0);
 }
 
 /**
@@ -1502,12 +1581,12 @@ static const char *directive_PythonInterpPerDirectory(cmd_parms *cmd,
 
 static const char *directive_PythonAutoReload(cmd_parms *cmd, 
                                               void *mconfig, int val) {
-    const char *rc = python_directive_flag(mconfig, "PythonAutoReload", val);
+    const char *rc = python_directive_flag(mconfig, "PythonAutoReload", val, 1);
 
     if (!rc) {
         py_config *conf = ap_get_module_config(cmd->server->module_config,
                                                &python_module);
-        return python_directive_flag(conf, "PythonAutoReload", val);
+        return python_directive_flag(conf, "PythonAutoReload", val, 1);
     }
     return rc;
 }
@@ -1523,15 +1602,28 @@ static const char *directive_PythonAutoReload(cmd_parms *cmd,
 static const char *directive_PythonOption(cmd_parms *cmd, void * mconfig, 
                                           const char *key, const char *val)
 {
-
     py_config *conf;
 
     conf = (py_config *) mconfig;
-    apr_table_set(conf->options, key, val);
 
-    conf = ap_get_module_config(cmd->server->module_config,
-                                &python_module);
-    apr_table_set(conf->options, key, val);
+    if(val!=NULL) {
+        apr_table_set(conf->options, key, val);
+
+        conf = ap_get_module_config(cmd->server->module_config,
+                                    &python_module);
+        apr_table_set(conf->options, key, val);
+    }
+    else {
+    	/** We don't remove the value, but set it
+    	    to an empty string. There is no possibility
+    	    of colliding with an actual value, since
+    	    an entry string precisely means 'remove the value' */
+        apr_table_set(conf->options, key, "");
+
+        conf = ap_get_module_config(cmd->server->module_config,
+                                    &python_module);
+        apr_table_set(conf->options, key, "");
+    }
 
     return NULL;
 }
@@ -1614,7 +1706,16 @@ static const char *directive_PythonHandlerModule(cmd_parms *cmd, void *mconfig,
     python_directive_handler(cmd, mconfig, "PythonLogHandler", val, SILENT);
     python_directive_handler(cmd, mconfig, "PythonCleanupHandler", val, SILENT);
 
+    /*
+     * XXX There is a bug here with PythonConnectionHandler which can
+     * cause an infinite loop when the handler is added to the handler
+     * list. Cause is unknown so simply disable it for now. If someone
+     * really needs a connection handler, they can use the directive
+     * PythonConnectionHandler explicitly still and not rely on the
+     * PythonHandlerModule directive doing it automatically.
+     *
     python_directive_handler(cmd, srv_conf, "PythonConnectionHandler", val, SILENT);
+    */
 
     return NULL;
 }
@@ -1811,6 +1912,7 @@ static void PythonChildInitHandler(apr_pool_t *p, server_rec *s)
 
         err:
             PyErr_Print();
+            fflush(stderr); 
             release_interpreter();
             return;
         }
@@ -1818,10 +1920,12 @@ static void PythonChildInitHandler(apr_pool_t *p, server_rec *s)
     success:
         /* now import the specified module */
         if (! PyImport_ImportModule((char *)module_name)) {
-            if (PyErr_Occurred())
+            if (PyErr_Occurred()) {
                 PyErr_Print();
+                fflush(stderr); 
+            }
             ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, s,
-                         "directive_PythonImport: error importing %s", module_name);
+                         "directive_PythonImport: error importing %s", (!module_name) ? "<null>" : module_name);
         }
 
         /* release interpreter */
@@ -2012,7 +2116,7 @@ command_rec python_commands[] =
     AP_INIT_FLAG(
         "PythonOptimize", directive_PythonOptimize, NULL, RSRC_CONF,
         "Set the equivalent of the -O command-line flag on the interpreter."),
-    AP_INIT_TAKE2(
+    AP_INIT_TAKE12(
         "PythonOption", directive_PythonOption, NULL, OR_ALL,
         "Useful to pass custom configuration information to scripts."),
     AP_INIT_TAKE1(
