@@ -18,7 +18,7 @@
  *
  * requestobject.c 
  *
- * $Id: requestobject.c 102649 2004-02-16 19:47:28Z grisha $
+ * $Id: requestobject.c 367839 2006-01-11 00:03:06Z jgallacher $
  *
  */
 
@@ -36,7 +36,7 @@ PyObject * MpRequest_FromRequest(request_rec *req)
 {
     requestobject *result;
 
-    result = PyObject_New(requestobject, &MpRequest_Type);
+    result = PyObject_GC_New(requestobject, &MpRequest_Type);
     if (! result)
         return PyErr_NoMemory();
 
@@ -62,10 +62,16 @@ PyObject * MpRequest_FromRequest(request_rec *req)
     result->rbuff = NULL;
     result->rbuff_pos = 0;
     result->rbuff_len = 0;
+    result->session = NULL;
+
+    // we make sure that the object dictionary is there
+    // before registering the object with the GC
+    PyObject_GC_Track(result);
 
     return (PyObject *) result;
 }
 
+
 /* Methods */
 
 /**
@@ -141,7 +147,7 @@ static PyObject *req_add_handler(requestobject *self, PyObject *args)
 
         /* then just append to hlist */
         hlist_append(self->request_rec->pool, self->hlo->head,
-                     handler, dir, 0);
+                     handler, dir, NOTSILENT);
     }
     else {
         /* this is a phase that we're not in */
@@ -157,11 +163,11 @@ static PyObject *req_add_handler(requestobject *self, PyObject *args)
         hle = apr_hash_get(req_config->dynhls, phase, APR_HASH_KEY_STRING);
 
         if (! hle) {
-            hle = hlist_new(self->request_rec->pool, handler, dir, 0);
+            hle = hlist_new(self->request_rec->pool, handler, dir, NOTSILENT);
             apr_hash_set(req_config->dynhls, phase, APR_HASH_KEY_STRING, hle);
         }
         else {
-            hlist_append(self->request_rec->pool, hle, handler, dir, 0);
+            hlist_append(self->request_rec->pool, hle, handler, dir, NOTSILENT);
         }
     }
     
@@ -346,7 +352,85 @@ static PyObject * req_get_options(requestobject *self, PyObject *args)
     py_config *conf =
         (py_config *) ap_get_module_config(self->request_rec->per_dir_config, 
                                            &python_module);
-    return MpTable_FromTable(conf->options);
+    apr_table_t* table = conf->options;
+
+    int i;
+    apr_array_header_t* ah = apr_table_elts(table);
+    apr_table_entry_t* elts = ah->elts;
+
+    /*
+     * We remove the empty values, since they cannot have been defined
+     * by the directive.
+     */
+    for(i=0;i<ah->nelts;i++,elts++) {
+        if(strlen(elts->val)==0) {
+            apr_table_unset(table,elts->key);
+        }
+    }
+
+    /* XXX shouldn't we free the apr_array_header_t* ah ? */
+
+    return MpTable_FromTable(table);
+}
+
+/**
+ ** request.get_session() 
+ **
+ *      Get the session instance, or create a new one.
+ *      This method just calls the python function
+ *      create_session() in mod_python/Session.py to actually create
+ *      the session instance.
+ */ 
+
+static PyObject *req_get_session(requestobject *self, PyObject *args)
+{
+#ifdef ENABLE_GET_SESSION
+    /* get_session is not ready for mod_python 3.2 release
+     * We'll leave the code intact but raise an error
+     */
+    PyObject *m;
+    PyObject *sid;
+    PyObject *req;
+    PyObject *result;
+
+    if (!self->session) {
+        /* Get the session id from the subprocess_env table if possible.
+         * If a session instance exists at the time of a req_internal_redirect()
+         * call, the session id will be added to the subprocess_env table with 
+         * the key PYSID.  req_internal_redirect() calls ap_internal_redirect(),
+         * which prepends all keys in subprocess_env with "REDIRECT_", so the key
+         * we want is REDIRECT_PYSID.
+         * req_internal_redirect() should have unlocked the session, so we do not
+         * need to worry about session deadlocking as a result of the redirect.
+         */ 
+        req = (PyObject *)self;
+        m = PyImport_ImportModule("mod_python.Session");
+        sid = PyObject_CallMethod(self->subprocess_env, "get", "(ss)","REDIRECT_PYSID", "");
+        self->session = PyObject_CallMethod(m, "create_session", "(OO)", req, sid);
+        Py_DECREF(m);
+        Py_DECREF(sid);
+
+        /* TODO
+         * Failure to DECREF req results in a serious memory leak.
+         * On the other hand, if the session created does not save 
+         * a reference to the request, we'll get a segfault.
+         * This should be documented for developers subclassing BaseSession.
+         * Maybe we should check self->session for the _req attribute
+         * and only do the decref if it exists? Or is there a better
+         * way to handle this? - Jim
+         */ 
+        Py_DECREF(req);
+        if (self->session == NULL)
+            return NULL;
+    }
+    result = self->session;
+    Py_INCREF(result);
+    return result;
+#else
+    PyErr_SetString(PyExc_NotImplementedError,
+                    "get_session is not implemented");
+    return NULL;
+#endif
 }
 
 /**
@@ -357,10 +441,59 @@ static PyObject * req_get_options(requestobject *self, PyObject *args)
 static PyObject * req_internal_redirect(requestobject *self, PyObject *args)
 {
     char *new_uri;
+#ifdef ENABLE_GET_SESSION
+    PyObject *sid;
+    PyObject *rc;
+#endif
 
     if (! PyArg_ParseTuple(args, "z", &new_uri))
         return NULL; /* error */
 
+#ifdef ENABLE_GET_SESSION
+    if (self->session){
+        /* A session exists, so save and unlock it before the redirect.
+         * The session instance is not preserved across the ap_internal_direct()
+         * call, so all we can do is save a reference to session id as a string
+         * and recreate the session instance when the redirected request
+         * is processed. This is handled by req_get_session().
+         * Note that ap_internal_redirect() makes a copy of the subprocess_env table,
+         * pre-pending all the keys in the table with "REDIRECT_".
+         * Each internal_redirect results in a save and a read, which may impact
+         * performance, depending on the type of persistent store used by the
+         * session class.
+         */
+
+        if (!(sid = PyObject_CallMethod(self->session, "id", NULL))) {
+            sid = PyString_FromString("");
+        }
+
+        /* Save the session id in the subprocess_env table. This id will be used
+         * to recreate the session instance when the redirected request is 
+         * processed.
+         */
+        if ((rc = PyObject_CallMethod(self->subprocess_env, "setdefault", "(sS)","PYSID", sid))) {
+            Py_DECREF(rc);
+        }
+        
+        Py_DECREF(sid);
+
+        /* Save the session instance so any session data can be restored when
+         * the redirected request is processed.
+         */
+        if (!(rc = PyObject_CallMethod(self->session, "save", NULL))) {
+            Py_DECREF(rc);
+        }
+
+        /* Unlock the session. The session instance will be recreated when
+         * the redirected request is processed. Failure to unlock the session
+         * here will result in a deadlock.
+         */
+        if (!(rc = PyObject_CallMethod(self->session, "unlock", NULL))) {
+            Py_DECREF(rc);
+        }
+    }
+#endif
+    
     Py_BEGIN_ALLOW_THREADS
     ap_internal_redirect(new_uri, self->request_rec);
     Py_END_ALLOW_THREADS
@@ -395,6 +528,17 @@ static PyObject * req_log_error(requestobject *self, PyObject *args)
     return Py_None;
 }
 
+ /**
+ ** request.meets_conditions(req self)
+ **
+ *  ap_meets_conditions wrapper
+ */
+
+static PyObject * req_meets_conditions(requestobject *self) {
+    return PyInt_FromLong(ap_meets_conditions(self->request_rec));
+}
+
+
 /**
  ** request.read(request self, int bytes)
  **
@@ -671,7 +815,7 @@ static PyObject *req_readlines(requestobject *self, PyObject *args)
         return PyErr_NoMemory();
 
     line = req_readline(self, rlargs);
-    while (line && !(strcmp(PyString_AsString(line), "") == 0)) {
+    while (line && (PyString_Size(line)>0)) {
         PyList_Append(result, line);
         size += PyString_Size(line);
         if ((sizehint != -1) && (size >= size))
@@ -880,7 +1024,7 @@ static PyObject * req_sendfile(requestobject *self, PyObject *args)
 
     Py_BEGIN_ALLOW_THREADS
     status=apr_stat(&finfo, fname,
-                    APR_READ, self->request_rec->pool);
+                    APR_FINFO_NORM, self->request_rec->pool);
     Py_END_ALLOW_THREADS
     if (status != APR_SUCCESS) {
         PyErr_SetString(PyExc_IOError, "Could not stat file for reading");
@@ -903,6 +1047,7 @@ static PyObject * req_sendfile(requestobject *self, PyObject *args)
         status = ap_send_fd(fd, self->request_rec, offset, 
                             len, &nbytes);
     Py_END_ALLOW_THREADS
+    apr_file_close(fd);
     
     if (status != APR_SUCCESS) {
         PyErr_SetString(PyExc_IOError, "Write failed, client closed connection.");
@@ -925,8 +1070,11 @@ static PyMethodDef request_methods[] = {
     {"get_config",            (PyCFunction) req_get_config,            METH_NOARGS},
     {"get_remote_host",       (PyCFunction) req_get_remote_host,       METH_VARARGS},
     {"get_options",           (PyCFunction) req_get_options,           METH_NOARGS},
+    {"get_session",           (PyCFunction) req_get_session,           METH_VARARGS},
+    {"write",                 (PyCFunction) req_write,                 METH_VARARGS},
     {"internal_redirect",     (PyCFunction) req_internal_redirect,     METH_VARARGS},
     {"log_error",             (PyCFunction) req_log_error,             METH_VARARGS},
+    {"meets_conditions",      (PyCFunction) req_meets_conditions,      METH_NOARGS},
     {"read",                  (PyCFunction) req_read,                  METH_VARARGS},
     {"readline",              (PyCFunction) req_readline,              METH_VARARGS},
     {"readlines",             (PyCFunction) req_readlines,             METH_VARARGS},
@@ -1088,6 +1236,15 @@ static int setreq_recmbr(requestobject *self, PyObject *val, void *name)
             apr_pstrdup(self->request_rec->pool, PyString_AsString(val));
         return 0;
     }
+    else if (strcmp(name, "path_info") == 0) {
+        if (! PyString_Check(val)) {
+            PyErr_SetString(PyExc_TypeError, "path_info must be a string");
+            return -1;
+        }
+        self->request_rec->path_info = 
+            apr_pstrdup(self->request_rec->pool, PyString_AsString(val));
+        return 0;
+    }
     
     return PyMember_SetOne((char*)self->request_rec, 
                            find_memberdef(request_rec_mbrs, (char*)name),
@@ -1139,7 +1296,7 @@ static PyObject *getreq_rec_ah(requestobject *self, void *name)
 {
     const PyMemberDef *md = find_memberdef(request_rec_mbrs, name);
     apr_array_header_t *ah = 
-        (apr_array_header_t *)((char *)self->request_rec + md->offset);
+        *(apr_array_header_t **)((char *)self->request_rec + md->offset);
 
     return tuple_from_array_header(ah);
 }
@@ -1154,7 +1311,7 @@ static PyObject *getreq_rec_ml(requestobject *self, void *name)
 {
     const PyMemberDef *md = find_memberdef(request_rec_mbrs, (char*)name);
     ap_method_list_t *ml = 
-        (ap_method_list_t *)((char *)self->request_rec + md->offset);
+        *(ap_method_list_t **)((char *)self->request_rec + md->offset);
 
     return tuple_from_method_list(ml);
 }
@@ -1200,28 +1357,35 @@ static PyObject *getmakeobj(requestobject* self, void *objname)
     PyObject *result = NULL;
 
     if (strcmp(name, "connection") == 0) {
-        if (!self->connection && self->request_rec->connection)
+        if (!self->connection && self->request_rec->connection) {
             self->connection = MpConn_FromConn(self->request_rec->connection);
+        }
         result = self->connection;
     }
     else if (strcmp(name, "server") == 0) {
-        if (!self->server && self->request_rec->server) 
+        if (!self->server && self->request_rec->server) {
             self->server = MpServer_FromServer(self->request_rec->server);
+        }
         result = self->server;
     }
     else if (strcmp(name, "next") == 0) {
-        if (!self->next && self->request_rec->next)
+        if (!self->next && self->request_rec->next) {
             self->next = MpRequest_FromRequest(self->request_rec->next);
+            ((requestobject*)self->next)->prev = self;
+        }
         result = self->next;
     }
     else if (strcmp(name, "prev") == 0) {
-        if (!self->prev && self->request_rec->prev)
+        if (!self->prev && self->request_rec->prev) {
             self->prev = MpRequest_FromRequest(self->request_rec->prev);
+            ((requestobject*)self->prev)->next = self;
+        }
         result = self->prev;
     }
     else if (strcmp(name, "main") == 0) {
-        if (!self->main && self->request_rec->main)
+        if (!self->main && self->request_rec->main) {
             self->main = MpRequest_FromRequest(self->request_rec->main);
+        }
         result = self->main;
     }
 
@@ -1258,7 +1422,6 @@ static PyGetSetDef request_getsets[] = {
     {"bytes_sent",   (getter)getreq_recmbr_off, NULL, "Bytes sent", "bytes_sent"},
     {"mtime",        (getter)getreq_recmbr_time, NULL, "Time resource was last modified", "mtime"},
     {"chunked",      (getter)getreq_recmbr, NULL, "Sending chunked transfer-coding", "chunked"},
-    {"boundary",     (getter)getreq_recmbr, NULL, "Multipart/byteranges boundary", "boundary"},
     {"range",        (getter)getreq_recmbr, NULL, "The Range: header", "range"},
     {"clength",      (getter)getreq_recmbr_off, NULL, "The \"real\" contenct length", "clength"},
     {"remaining",    (getter)getreq_recmbr_off, NULL, "Bytes left to read", "remaining"},
@@ -1279,7 +1442,7 @@ static PyGetSetDef request_getsets[] = {
     {"uri",           (getter)getreq_recmbr, NULL, "The path portion of URI", "uri"},
     {"filename",      (getter)getreq_recmbr, (setter)setreq_recmbr, "The file name on disk that this request corresponds to", "filename"},
     {"canonical_filename", (getter)getreq_recmbr, NULL, "The true filename (req.filename is canonicalized if they dont match)", "canonical_filename"},
-    {"path_info",     (getter)getreq_recmbr, NULL, "Path_info, if any", "path_info"},
+    {"path_info",     (getter)getreq_recmbr, (setter)setreq_recmbr, "Path_info, if any", "path_info"},
     {"args",          (getter)getreq_recmbr, NULL, "QUERY_ARGS, if any", "args"},
     {"finfo",         (getter)getreq_rec_fi, NULL, "File information", "finfo"},
     {"parsed_uri",    (getter)getreq_rec_uri, NULL, "Components of URI", "parsed_uri"},
@@ -1309,31 +1472,77 @@ static struct PyMemberDef request_members[] = {
     {NULL}  /* Sentinel */
 };
 
+/**
+ ** request_tp_clear
+ **
+ *    
+ */
+static int request_tp_clear(requestobject *self)
+{   
+    PyObject* tmp;
+    /* TODO: create a macro for the following repetitive code */ 
+    tmp=self->dict; self->dict=NULL; Py_XDECREF(tmp);
+    tmp=self->connection; self->connection=NULL; Py_XDECREF(tmp);
+    tmp=self->server; self->server=NULL; Py_XDECREF(tmp);
+    tmp=self->next; self->next=NULL; Py_XDECREF(tmp);
+    tmp=self->prev; self->prev=NULL; Py_XDECREF(tmp);
+    tmp=self->main; self->main=NULL; Py_XDECREF(tmp);
+    tmp=self->headers_in; self->headers_in=NULL; Py_XDECREF(tmp);
+    tmp=self->headers_out; self->headers_out=NULL; Py_XDECREF(tmp);
+    tmp=self->err_headers_out; self->err_headers_out=NULL; Py_XDECREF(tmp);
+    tmp=self->subprocess_env; self->subprocess_env=NULL; Py_XDECREF(tmp);
+    tmp=self->notes; self->notes=NULL; Py_XDECREF(tmp);
+    tmp=self->phase; self->phase=NULL; Py_XDECREF(tmp);
+    tmp=self->hlo; self->hlo=NULL; Py_XDECREF(tmp);
+    tmp=self->session; self->session=NULL; Py_XDECREF(tmp);
+    
+    return 0;
+}
+
+
 /**
  ** request_dealloc
  **
  *
  */
 
-static void request_dealloc(requestobject *self)
+static void request_tp_dealloc(requestobject *self)
 {  
-    Py_XDECREF(self->dict);
-    Py_XDECREF(self->connection);
-    Py_XDECREF(self->server);
-    Py_XDECREF(self->next);
-    Py_XDECREF(self->prev);
-    Py_XDECREF(self->main);
-    Py_XDECREF(self->headers_in);
-    Py_XDECREF(self->headers_out);
-    Py_XDECREF(self->err_headers_out);
-    Py_XDECREF(self->subprocess_env);
-    Py_XDECREF(self->notes);
-    Py_XDECREF(self->phase);
-    Py_XDECREF(self->hlo);
+    // de-register the object from the GC
+    // before its deallocation, to prevent the
+    // GC to run on a partially de-allocated object
+    PyObject_GC_UnTrack(self);
+    
+    request_tp_clear(self);
 
-    PyObject_Del(self);
+    PyObject_GC_Del(self);
 }
 
+/**
+ ** request_tp_traverse
+ **
+ *    Traversal of the request object
+ */
+static int request_tp_traverse(requestobject* self, visitproc visit, void *arg) {
+    int result;
+
+    if(self->dict) {result = visit(self->dict,arg); if(result) return result;}
+    if(self->connection) {result = visit(self->connection,arg); if(result) return result;}
+    if(self->server) {result = visit(self->server,arg); if(result) return result;}
+    if(self->next) {result = visit(self->next,arg); if(result) return result;}
+    if(self->prev) {result = visit(self->prev,arg); if(result) return result;}
+    if(self->main) {result = visit(self->main,arg); if(result) return result;}
+    if(self->headers_in) {result = visit(self->headers_in,arg); if(result) return result;}
+    if(self->headers_out) {result = visit(self->headers_out,arg); if(result) return result;}
+    if(self->err_headers_out) {result = visit(self->err_headers_out,arg); if(result) return result;}
+    if(self->subprocess_env) {result = visit(self->subprocess_env,arg); if(result) return result;}
+    if(self->notes) {result = visit(self->notes,arg); if(result) return result;}
+    if(self->phase) {result = visit(self->phase,arg); if(result) return result;}
+    if(self->session) {result = visit(self->session,arg); if(result) return result;}
+    
+    // no need to Py_DECREF(dict) since the reference is borrowed
+    return 0;
+}
 static char request_doc[] =
 "Apache request_rec structure\n";
 
@@ -1343,42 +1552,43 @@ PyTypeObject MpRequest_Type = {
     "mp_request",
     sizeof(requestobject),
     0,
-    (destructor) request_dealloc,    /*tp_dealloc*/
-    0,                               /*tp_print*/
-    0,                               /*tp_getattr*/
-    0,                               /*tp_setattr*/
-    0,                               /*tp_compare*/
-    0,                               /*tp_repr*/
-    0,                               /*tp_as_number*/
-    0,                               /*tp_as_sequence*/
-    0,                               /*tp_as_mapping*/
-    0,                               /*tp_hash*/
-    0,                               /* tp_call */
-    0,                               /* tp_str */
-    PyObject_GenericGetAttr,         /* tp_getattro */
-    PyObject_GenericSetAttr,         /* tp_setattro */
-    0,                               /* tp_as_buffer */
-    Py_TPFLAGS_DEFAULT |
-    Py_TPFLAGS_BASETYPE,             /* tp_flags */
-    request_doc,                     /* tp_doc */
-    0,                               /* tp_traverse */
-    0,                               /* tp_clear */
-    0,                               /* tp_richcompare */
-    0,                               /* tp_weaklistoffset */
-    0,                               /* tp_iter */
-    0,                               /* tp_iternext */
-    request_methods,                 /* tp_methods */
-    request_members,                 /* tp_members */
-    request_getsets,                 /* tp_getset */
-    0,                               /* tp_base */
-    0,                               /* tp_dict */
-    0,                               /* tp_descr_get */
-    0,                               /* tp_descr_set */
-    offsetof(requestobject, dict),   /* tp_dictoffset */
-    0,                               /* tp_init */
-    0,                               /* tp_alloc */
-    0,                               /* tp_new */
-    (destructor)request_dealloc,     /* tp_free */
+    (destructor)request_tp_dealloc,       /*tp_dealloc*/
+    0,                                 /*tp_print*/
+    0,                                 /*tp_getattr*/
+    0,                                 /*tp_setattr*/
+    0,                                 /*tp_compare*/
+    0,                                 /*tp_repr*/
+    0,                                 /*tp_as_number*/
+    0,                                 /*tp_as_sequence*/
+    0,                                 /*tp_as_mapping*/
+    0,                                 /*tp_hash*/
+    0,                                 /* tp_call */
+    0,                                 /* tp_str */
+    PyObject_GenericGetAttr,           /* tp_getattro */
+    PyObject_GenericSetAttr,           /* tp_setattro */
+    0,                                 /* tp_as_buffer */
+    Py_TPFLAGS_DEFAULT |  
+    Py_TPFLAGS_BASETYPE|  
+    Py_TPFLAGS_HAVE_GC ,               /* tp_flags */
+    request_doc,                       /* tp_doc */
+    (traverseproc)request_tp_traverse, /* tp_traverse */
+    (inquiry)request_tp_clear,         /* tp_clear */
+    0,                                 /* tp_richcompare */
+    0,                                 /* tp_weaklistoffset */
+    0,                                 /* tp_iter */
+    0,                                 /* tp_iternext */
+    request_methods,                   /* tp_methods */
+    request_members,                   /* tp_members */
+    request_getsets,                   /* tp_getset */
+    0,                                 /* tp_base */
+    0,                                 /* tp_dict */
+    0,                                 /* tp_descr_get */
+    0,                                 /* tp_descr_set */
+    offsetof(requestobject, dict),     /* tp_dictoffset */
+    0,                                 /* tp_init */
+    0,                                 /* tp_alloc */
+    0,                                 /* tp_new */
+    0,                                 /* tp_free */
 };
 
 
